"""iniverse testor"""


# from hai.testor.test_module import TestModule


class Testor(object):
    def __init__(self) -> None:
        
        pass

    def __call__(self):
        pass

    def test_module(self, module_name, **kwargs):
        """entry function for algorithm module test"""
        from .test_module import TestModule
        t = TestModule()
        t(model_name=module_name, **kwargs)


import os, sys
import time
import damei as dm
try:
    import cv2
except:
    pass
import shutil

from ..utils.config_loader import PyConfigLoader
from ..registry import MODULES, SCRIPTS, IOS
from .dataloader import Dataloader

logger = dm.getLogger('uaii loaders')


# @MODULES.register_module(name='visible_loader')
class VisibleLoader(Dataloader):
    name = 'visible_loader'
    status = 'stopped'
    description = '可见光数据加载器'
    default_cfg = None
    tag = 'cloud'

    def __init__(self, cfg=None):
        if isinstance(cfg, str) and cfg:
            cfg = PyConfigLoader(cfg)
        self.cfg = cfg
        source = cfg.get('source', None)
        imgsz = cfg.get('img_size', 640)
        assert source
        super(VisibleLoader, self).__init__(source=source, imgsz=imgsz)
        self.mi = None
        self.mo = None
        self.que_wait = self.cfg.output_stream.que.get('wait', True)

    def __call__(self, *args, **kwargs):
        """不断加载数据，存到que中"""
        wait = kwargs.pop('wait', self.que_wait)  # 等待有后面的东西拿掉再推

        for img_idx, data in enumerate(self.dataset):
            paths, imgs, im0s, vid_cap = data
            if wait:
                while True:
                    if len(self.mo.que) == 0:
                        self.mo.push([paths, imgs, im0s, vid_cap])
                        break
                    time.sleep(0.0001)
            else:
                self.mo.push([paths, imgs, im0s, vid_cap])
                time.sleep(0.0001)  # 0.1 ms
            # print(imgs.shape)
        # print(self.mo)


class XsensingLoader(object):
    """Xsensing传感器加载数据"""

    def __init__(self, ip, port, imgsz=640):
        self.ip = ip
        self.port = port
        self.imgsz = imgsz
        self.dataloader = None

    def __call__(self, show=True, raw=True, scale=1):
        ip = self.ip
        port = self.port
        imgsz = self.imgsz
        rtsp = f"rtsp://{ip}:{port}/live/stream"
        dataloader = Dataloader(source=rtsp, imgsz=imgsz)

        for i, (path, img, im0, cap) in enumerate(dataloader.dataset):
            # print(i, path, img.shape, type(img), cap)

            raw_img = im0[0]
            if raw_img is None:
                continue
            # print(raw_img.shape)
            resized_img = img[0].transpose(1, 2, 0)
            # print(resized_img, resized_img.shape)
            if show:
                show_img = raw_img if raw else resized_img
                h, w, c = show_img.shape
                show_img = cv2.resize(show_img, (int(w * scale), int(h * scale))) if scale else show_img
                cv2.imshow('img', show_img)
                if cv2.waitKey(5) == ord('q'):  # q to quit
                    raise StopIteration

    def show(self, show=True, raw=True, scale=1):
        self.__call__(show=show, raw=raw, scale=scale)

    def record(self, dir, seconds=60):
        """
        :param dir: saved directory
        :param seconds: record time, default 60s
        :return:
        """
        dir = os.path.abspath(dir)
        if os.path.exists(dir):
            shutil.rmtree(dir)
        os.makedirs(dir)

        save_log = []

        rtsp = f"rtsp://{self.ip}:{self.port}/live/stream"  # 目前只支持1个流
        dataloader = Dataloader(source=rtsp, imgsz=self.imgsz)
        save_log.append(f'Source: {rtsp}\n')
        et = dm.plus_time(seconds=seconds)

        for idx, (path, img, im0, cap) in enumerate(dataloader.dataset):
            raw_img = im0[0]
            if raw_img is None:
                continue
            ct = dm.current_time()
            if not dm.within_time(et, ct):
                break

            file_path = f'{dir}/{idx:0>8}.jpg'
            logger.info(f'Saving img: {file_path}')
            cv2.imwrite(filename=file_path, img=raw_img)
            info = f'[{ct}] {dir}/{idx:0>8}.jpg {raw_img.shape}\n'
            save_log.append(info)

        with open(f'{dir}/record.log', 'w') as f:
            f.writelines(save_log)






class BaseWorkerModel:
    """
    用于被控制工人模型的基类，提供了面向切片的自动流式输出等功能
    Usage:
        class MyWorker(WorkerBaseModel):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
            
            @WorkerBaseModel.auto_stream  # 自动流式输出
            def inference(self, **kwargs):
                “”“重写推理函数”“”
                return "output"
            
        worker = MyWorker()
        worker.start()
    """
    def __init__(self, **kwargs) -> None:
        self.name = kwargs.pop('name', "hepai/worker-base-model")
        self.trainable = kwargs.pop('trainable', False)
        self.inferable = kwargs.pop('inferable', False)

    @staticmethod
    def convert(ret):
        for x in ret:
            yield x

    @staticmethod
    def auto_stream(func):
        """
        自动流式输出
        """
        def warpper(*args, **kwargs):
            stream = kwargs.get('stream', False)    
            output = func(*args, **kwargs)
            if stream:
                is_generator =  type(output) == type((i for i in range(10)))
                if isinstance(output, str):
                    return BaseWorkerModel.convert(output)
                elif is_generator:
                    return output
                else:
                    raise TypeError(f"Output type {type(output)} is supported in stream mode, please use 'str' type.")
            else:
                return output
        return warpper
    
    def inference(self, **kwargs):
        """需要重写函数"""
        raise NotImplementedError("Please implement this method")
    
    def train(self, **kwargs):
        """需要重写函数"""
        raise NotImplementedError("Please implement this method")
    
    def evaluate(self, **kwargs):
        """需要重写函数"""
        raise NotImplementedError("Please implement this method")
    

    def get_misc(self, **kwargs):
        """需要重写函数"""
        misc = dict()
        misc['trainable'] = self.trainable
        misc['inferable'] = self.inferable
        return misc


import functools
import eventlet
import socketio
import logging
import damei

logger = damei.getLogger('uais_socket')


class SocketApp(object):
    debug = False
    sio = socketio.Server(logger=debug, engineio_logger=debug)
    app = socketio.WSGIApp(sio, static_files={
        '/': {'content_type': 'text/html', 'filename': 'index.html'}})
    uaii = None

    @sio.event
    def connect(sid, environ):
        logger.info(f'客户端连接，sid: {sid}.')
        # print(f'Client connected, sid: {sid}')
        # logger.info(f'Client connected, sid: {sid}')
        # print('environ', environ)

    @sio.event
    def my_message(sid, data):
        print(f'my_message received message: {data}')

    @sio.event
    def server_run_func(sid, data):
        logger.info(f'server run_func, data: {data}')
        # print(f'server run_func, demo_for_dm.data: {demo_for_dm.data}')
        return 'dame', 'dame2'

    @sio.event
    def disconnect(sid):
        logger.info(f'客户端断开连接，sid: {sid}.')
        # print(f'Client disconnected, sid: {sid}')

    @sio.event
    def ps(sid, *args, **kwargs):
        """参数会存在args[0]里"""
        params = args[0]
        # print('客户端让我读取ps信息')
        # params = kwargs.get('params')
        logger.info(f'客户端请求读取ps信息. '
                    # f'sid: {sid} args: {args} kwargs: {kwargs}'
                    )
        ret = SocketApp.uaii.ps()

        return ret

    @sio.event
    def uaii(self, *args):
        """处理来自客户端的uaii请求"""
        print(args)
        func, args, kwargs = args[0][0]
        logger.info(f'收到请求. func: {func} args: {args} **kwargs: {kwargs}')
        if not hasattr(SocketApp.uaii, func):
            logger.error(f'uaii不具有函数：{func}')
        func = getattr(SocketApp.uaii, func)
        # logger.info(f'func: {func} {type(func)}')
        ret = func(*args, **kwargs)

        return ret


def create_socket_app():
    socketapp = SocketApp()
    return socketapp


if __name__ == '__main__':
    app = create_socket_app()
    eventlet.wsgi.server(eventlet.listen(('', 5000)), app)



# -*- coding=utf-8 -*-
import base64
import hashlib
import logging
import socket
import json
import platform
import time
import hmac

try:
    import ssl
except ImportError:
    ssl = None

from multiprocessing import Process, Manager, Queue, pool
from threading import RLock, Thread

try:
    # python3.6
    from http import HTTPStatus
    from urllib.request import Request, urlopen, ProxyHandler, HTTPSHandler, build_opener
    from urllib.parse import urlencode, unquote_plus, quote
    from urllib.error import HTTPError, URLError
except ImportError:
    # python2.7
    import httplib as HTTPStatus
    from urllib2 import Request, urlopen, HTTPError, URLError, ProxyHandler, HTTPSHandler, build_opener
    from urllib import urlencode, unquote_plus, quote

    base64.encodebytes = base64.encodestring

from .commons import synchronized_with_attr, truncate, python_version_bellow
from .params import group_key, parse_key, is_valid
from .files import read_file_str, save_file, delete_file
from .exception import NacosException, NacosRequestException
from .listener import Event, SimpleListenerManager
from .timer import NacosTimer, NacosTimerManager

logging.basicConfig()
logger = logging.getLogger(__name__)

DEBUG = False
VERSION = "0.1.7"

DEFAULT_GROUP_NAME = "DEFAULT_GROUP"
DEFAULT_NAMESPACE = ""

WORD_SEPARATOR = u'\x02'
LINE_SEPARATOR = u'\x01'

DEFAULTS = {
    "APP_NAME": "Nacos-SDK-Python",
    "TIMEOUT": 3,  # in seconds
    "PULLING_TIMEOUT": 30,  # in seconds
    "PULLING_CONFIG_SIZE": 3000,
    "CALLBACK_THREAD_NUM": 10,
    "FAILOVER_BASE": "nacos-data/data",
    "SNAPSHOT_BASE": "nacos-data/snapshot",
}

OPTIONS = {"default_timeout", "pulling_timeout", "pulling_config_size", "callback_thread_num", "failover_base",
           "snapshot_base", "no_snapshot", "proxies"}


def process_common_config_params(data_id, group):
    if not group or not group.strip():
        group = DEFAULT_GROUP_NAME
    else:
        group = group.strip()

    if not data_id or not is_valid(data_id):
        raise NacosException("Invalid dataId.")

    if not is_valid(group):
        raise NacosException("Invalid group.")
    return data_id, group


def parse_pulling_result(result):
    if not result:
        return list()
    ret = list()
    for i in unquote_plus(result.decode()).split(LINE_SEPARATOR):
        if not i.strip():
            continue
        sp = i.split(WORD_SEPARATOR)
        if len(sp) < 3:
            sp.append("")
        ret.append(sp)
    return ret


class WatcherWrap:
    def __init__(self, key, callback, last_md5=None):
        self.callback = callback
        self.last_md5 = last_md5
        self.watch_key = key


class CacheData:
    def __init__(self, key, client):
        self.key = key
        local_value = read_file_str(client.failover_base, key) or read_file_str(client.snapshot_base, key)
        self.content = local_value
        self.md5 = hashlib.md5(local_value.encode("UTF-8")).hexdigest() if local_value else None
        self.is_init = True
        if not self.md5:
            logger.debug("[init-cache] cache for %s does not have local value" % key)


class SubscribedLocalInstance(object):
    def __init__(self, key, instance):
        self.key = key
        self.instance_id = instance["instanceId"]
        self.md5 = NacosClient.get_md5(str(instance))
        self.instance = instance


class SubscribedLocalManager(object):
    def __init__(self):
        self.manager = {
            # "key1": {
            #     "LOCAL_INSTANCES": {
            #         "instanceId1": None,
            #         "instanceId2": None,
            #         "instanceId3": None,
            #         "instanceId4": None
            #     },
            #     "LISTENER_MANAGER": None
            # },
            # "key2": {
            #     "LOCAL_INSTANCES": {
            #         "instanceId1": "",
            #         "instanceId2": "",
            #         "instanceId3": "",
            #         "instanceId4": ""
            #     },
            #     "LISTENER_MANAGER": None
            # }
        }

    def do_listener_launch(self, key, event, slc):
        listener_manager = self.get_local_listener_manager(key)
        if listener_manager and isinstance(listener_manager, SimpleListenerManager):
            listener_manager.do_launch(event, slc)

    def get_local_listener_manager(self, key):
        key_node = self.manager.get(key)
        if not key_node:
            return None
        return key_node.get("LISTENER_MANAGER")

    def add_local_listener(self, key, listener_fn):
        if not self.manager.get(key):
            self.manager[key] = {}
        local_listener_manager = self.manager.get(key).get("LISTENER_MANAGER")

        if not local_listener_manager or not isinstance(local_listener_manager, SimpleListenerManager):
            self.manager.get(key)["LISTENER_MANAGER"] = SimpleListenerManager()
        local_listener_manager = self.manager.get(key).get("LISTENER_MANAGER")
        if not local_listener_manager:
            return self
        if isinstance(listener_fn, list):
            listener_fn = tuple(listener_fn)
            local_listener_manager.add_listeners(*listener_fn)
        if isinstance(listener_fn, tuple):
            local_listener_manager.add_listeners(*listener_fn)
        #  just single listener function
        else:
            local_listener_manager.add_listener(listener_fn)
        return self

    def add_local_listener_manager(self, key, listener_manager):
        key_node = self.manager.get(key)
        if key_node is None:
            key_node = {}
        key_node["LISTENER_MANAGER"] = listener_manager
        return self

    def get_local_instances(self, key):
        if not self.manager.get(key):
            return None
        return self.manager.get(key).get("LOCAL_INSTANCES")

    def add_local_instance(self, slc):
        if not self.manager.get(slc.key):
            self.manager[slc.key] = {}
        if not self.manager.get(slc.key).get('LOCAL_INSTANCES'):
            self.manager.get(slc.key)['LOCAL_INSTANCES'] = {}
        self.manager.get(slc.key)['LOCAL_INSTANCES'][slc.instance_id] = slc
        return self

    def remove_local_instance(self, slc):
        key_node = self.manager.get(slc.key)
        if not key_node:
            return self
        local_instances_node = key_node.get("LOCAL_INSTANCES")
        if not local_instances_node:
            return self
        local_instance = local_instances_node.get(slc.instance_id)
        if not local_instance:
            return self
        local_instances_node.pop(slc.instance_id)
        return self


def parse_nacos_server_addr(server_addr):
    sp = server_addr.split(":")
    if len(sp) == 3:
        return sp[0] + ":" + sp[1], int(sp[2])
    else:
        port = int(sp[1]) if len(sp) > 1 else 8848
        return sp[0], port


class NacosClient:
    debug = False

    @staticmethod
    def set_debugging():
        if not NacosClient.debug:
            global logger
            logger = logging.getLogger("nacos")
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s:%(message)s"))
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)
            NacosClient.debug = True

    @staticmethod
    def get_md5(content):
        return hashlib.md5(content.encode("UTF-8")).hexdigest() if content is not None else None

    def __init__(self, server_addresses, endpoint=None, namespace=None, ak=None, sk=None, username=None, password=None):
        self.server_list = list()

        try:
            for server_addr in server_addresses.split(","):
                self.server_list.append(parse_nacos_server_addr(server_addr.strip()))
        except Exception as ex:
            logger.exception("[init] bad server address for %s" % server_addresses)
            raise ex

        self.current_server = self.server_list[0]

        self.endpoint = endpoint
        self.namespace = namespace or DEFAULT_NAMESPACE or ""
        self.ak = ak
        self.sk = sk
        self.username = username
        self.password = password

        self.server_list_lock = RLock()
        self.server_offset = 0

        self.watcher_mapping = dict()
        self.subscribed_local_manager = SubscribedLocalManager()
        self.subscribe_timer_manager = NacosTimerManager()
        self.pulling_lock = RLock()
        self.puller_mapping = None
        self.notify_queue = None
        self.callback_tread_pool = None
        self.process_mgr = None

        self.default_timeout = DEFAULTS["TIMEOUT"]
        self.auth_enabled = self.ak and self.sk
        self.cai_enabled = True
        self.pulling_timeout = DEFAULTS["PULLING_TIMEOUT"]
        self.pulling_config_size = DEFAULTS["PULLING_CONFIG_SIZE"]
        self.callback_thread_num = DEFAULTS["CALLBACK_THREAD_NUM"]
        self.failover_base = DEFAULTS["FAILOVER_BASE"]
        self.snapshot_base = DEFAULTS["SNAPSHOT_BASE"]
        self.no_snapshot = False
        self.proxies = None

        logger.info("[client-init] endpoint:%s, tenant:%s" % (endpoint, namespace))

    def set_options(self, **kwargs):
        for k, v in kwargs.items():
            if k not in OPTIONS:
                logger.warning("[set_options] unknown option:%s, ignored" % k)
                continue

            logger.debug("[set_options] key:%s, value:%s" % (k, v))
            setattr(self, k, v)

    def change_server(self):
        with self.server_list_lock:
            self.server_offset = (self.server_offset + 1) % len(self.server_list)
            self.current_server = self.server_list[self.server_offset]

    def get_server(self):
        logger.info("[get-server] use server:%s" % str(self.current_server))
        return self.current_server

    def remove_config(self, data_id, group, timeout=None):
        data_id, group = process_common_config_params(data_id, group)
        logger.info(
            "[remove] data_id:%s, group:%s, namespace:%s, timeout:%s" % (data_id, group, self.namespace, timeout))

        params = {
            "dataId": data_id,
            "group": group,
        }
        if self.namespace:
            params["tenant"] = self.namespace

        try:
            resp = self._do_sync_req("/nacos/v1/cs/configs", None, None, params,
                                     timeout or self.default_timeout, "DELETE")
            c = resp.read()
            logger.info("[remove] remove group:%s, data_id:%s, server response:%s" % (
                group, data_id, c))
            return c == b"true"
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                logger.error(
                    "[remove] no right for namespace:%s, group:%s, data_id:%s" % (self.namespace, group, data_id))
                raise NacosException("Insufficient privilege.")
            else:
                logger.error("[remove] error code [:%s] for namespace:%s, group:%s, data_id:%s" % (
                    e.code, self.namespace, group, data_id))
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[remove] exception %s occur" % str(e))
            raise

    def publish_config(self, data_id, group, content, app_name=None, config_type=None, timeout=None):
        if content is None:
            raise NacosException("Can not publish none content, use remove instead.")

        data_id, group = process_common_config_params(data_id, group)
        if type(content) == bytes:
            content = content.decode("UTF-8")

        logger.info("[publish] data_id:%s, group:%s, namespace:%s, content:%s, timeout:%s" % (
            data_id, group, self.namespace, truncate(content), timeout))

        params = {
            "dataId": data_id,
            "group": group,
            "content": content.encode("UTF-8"),
        }

        if self.namespace:
            params["tenant"] = self.namespace

        if app_name:
            params["appName"] = app_name

        if config_type:
            params["type"] = config_type

        try:
            resp = self._do_sync_req("/nacos/v1/cs/configs", None, None, params,
                                     timeout or self.default_timeout, "POST")
            c = resp.read()
            logger.info("[publish] publish content, group:%s, data_id:%s, server response:%s" % (
                group, data_id, c))
            return c == b"true"
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                raise NacosException("Insufficient privilege.")
            else:
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[publish] exception %s occur" % str(e))
            raise

    def get_config(self, data_id, group, timeout=None, no_snapshot=None):
        no_snapshot = self.no_snapshot if no_snapshot is None else no_snapshot
        data_id, group = process_common_config_params(data_id, group)
        logger.info("[get-config] data_id:%s, group:%s, namespace:%s, timeout:%s" % (
            data_id, group, self.namespace, timeout))

        params = {
            "dataId": data_id,
            "group": group,
        }
        if self.namespace:
            params["tenant"] = self.namespace

        cache_key = group_key(data_id, group, self.namespace)
        # get from failover
        content = read_file_str(self.failover_base, cache_key)
        if content is None:
            logger.debug("[get-config] failover config is not exist for %s, try to get from server" % cache_key)
        else:
            logger.debug("[get-config] get %s from failover directory, content is %s" % (cache_key, truncate(content)))
            return content

        # get from server
        try:
            resp = self._do_sync_req("/nacos/v1/cs/configs", None, params, None, timeout or self.default_timeout)
            content = resp.read().decode("UTF-8")
        except HTTPError as e:
            if e.code == HTTPStatus.NOT_FOUND:
                logger.warning(
                    "[get-config] config not found for data_id:%s, group:%s, namespace:%s, try to delete snapshot" % (
                        data_id, group, self.namespace))
                delete_file(self.snapshot_base, cache_key)
                return None
            elif e.code == HTTPStatus.CONFLICT:
                logger.error(
                    "[get-config] config being modified concurrently for data_id:%s, group:%s, namespace:%s" % (
                        data_id, group, self.namespace))
            elif e.code == HTTPStatus.FORBIDDEN:
                logger.error("[get-config] no right for data_id:%s, group:%s, namespace:%s" % (
                    data_id, group, self.namespace))
                raise NacosException("Insufficient privilege.")
            else:
                logger.error("[get-config] error code [:%s] for data_id:%s, group:%s, namespace:%s" % (
                    e.code, data_id, group, self.namespace))
                if no_snapshot:
                    raise
        except Exception as e:
            logger.exception("[get-config] exception %s occur" % str(e))
            if no_snapshot:
                raise

        if no_snapshot:
            return content

        if content is not None:
            logger.info(
                "[get-config] content from server:%s, data_id:%s, group:%s, namespace:%s, try to save snapshot" % (
                    truncate(content), data_id, group, self.namespace))
            try:
                save_file(self.snapshot_base, cache_key, content)
            except Exception as e:
                logger.exception("[get-config] save snapshot failed for %s, data_id:%s, group:%s, namespace:%s" % (
                    data_id, group, self.namespace, str(e)))
            return content

        logger.error("[get-config] get config from server failed, try snapshot, data_id:%s, group:%s, namespace:%s" % (
            data_id, group, self.namespace))
        content = read_file_str(self.snapshot_base, cache_key)
        if content is None:
            logger.warning("[get-config] snapshot is not exist for %s." % cache_key)
        else:
            logger.debug("[get-config] get %s from snapshot directory, content is %s" % (cache_key, truncate(content)))
            return content

    def get_configs(self, timeout=None, no_snapshot=None, group="", page_no=1, page_size=1000):
        no_snapshot = self.no_snapshot if no_snapshot is None else no_snapshot
        logger.info("[get-configs] namespace:%s, timeout:%s, group:%s, page_no:%s, page_size:%s" % (
            self.namespace, timeout, group, page_no, page_size))

        params = {
            "dataId": "",
            "group": group,
            "search": "accurate",
            "pageNo": page_no,
            "pageSize": page_size,
        }
        if self.namespace:
            params["tenant"] = self.namespace

        cache_key = group_key("", "", self.namespace)
        # get from failover
        content = read_file_str(self.failover_base, cache_key)
        if content is None:
            logger.debug("[get-config] failover config is not exist for %s, try to get from server" % cache_key)
        else:
            logger.debug("[get-config] get %s from failover directory, content is %s" % (cache_key, truncate(content)))
            return json.loads(content)

        # get from server
        try:
            resp = self._do_sync_req("/nacos/v1/cs/configs", None, params, None, timeout or self.default_timeout)
            content = resp.read().decode("UTF-8")
        except HTTPError as e:
            if e.code == HTTPStatus.CONFLICT:
                logger.error(
                    "[get-configs] configs being modified concurrently for namespace:%s" % self.namespace)
            elif e.code == HTTPStatus.FORBIDDEN:
                logger.error("[get-configs] no right for namespace:%s" % self.namespace)
                raise NacosException("Insufficient privilege.")
            else:
                logger.error("[get-configs] error code [:%s] for namespace:%s" % (e.code, self.namespace))
                if no_snapshot:
                    raise
        except Exception as e:
            logger.exception("[get-config] exception %s occur" % str(e))
            if no_snapshot:
                raise

        if no_snapshot:
            return json.loads(content)

        if content is not None:
            logger.info(
                "[get-configs] content from server:%s, namespace:%s, try to save snapshot" % (
                    truncate(content), self.namespace))
            try:
                save_file(self.snapshot_base, cache_key, content)

                for item in json.loads(content).get("pageItems"):
                    data_id = item.get('dataId')
                    group = item.get('group')
                    item_content = item.get('content')
                    item_cache_key = group_key(data_id, group, self.namespace)
                    save_file(self.snapshot_base, item_cache_key, item_content)
            except Exception as e:
                logger.exception("[get-configs] save snapshot failed for %s, namespace:%s" % (
                    str(e), self.namespace))
            return json.loads(content)

        logger.error("[get-configs] get config from server failed, try snapshot, namespace:%s" % self.namespace)
        content = read_file_str(self.snapshot_base, cache_key)
        if content is None:
            logger.warning("[get-configs] snapshot is not exist for %s." % cache_key)
        else:
            logger.debug("[get-configs] get %s from snapshot directory, content is %s" % (cache_key, truncate(content)))
            return json.loads(content)

    @synchronized_with_attr("pulling_lock")
    def add_config_watcher(self, data_id, group, cb, content=None):
        self.add_config_watchers(data_id, group, [cb], content)

    @synchronized_with_attr("pulling_lock")
    def add_config_watchers(self, data_id, group, cb_list, content=None):
        if not cb_list:
            raise NacosException("A callback function is needed.")
        data_id, group = process_common_config_params(data_id, group)
        logger.info("[add-watcher] data_id:%s, group:%s, namespace:%s" % (data_id, group, self.namespace))
        cache_key = group_key(data_id, group, self.namespace)
        wl = self.watcher_mapping.get(cache_key)
        if not wl:
            wl = list()
            self.watcher_mapping[cache_key] = wl
        if not content:
            content = self.get_config(data_id, group)
        last_md5 = NacosClient.get_md5(content)
        for cb in cb_list:
            wl.append(WatcherWrap(cache_key, cb, last_md5))
            logger.info("[add-watcher] watcher has been added for key:%s, new callback is:%s, callback number is:%s" % (
                cache_key, cb.__name__, len(wl)))

        if self.puller_mapping is None:
            logger.debug("[add-watcher] pulling should be initialized")
            self._init_pulling()

        if cache_key in self.puller_mapping:
            logger.debug("[add-watcher] key:%s is already in pulling" % cache_key)
            return

        for key, puller_info in self.puller_mapping.items():
            if len(puller_info[1]) < self.pulling_config_size:
                logger.debug("[add-watcher] puller:%s is available, add key:%s" % (puller_info[0], cache_key))
                puller_info[1].append(cache_key)
                self.puller_mapping[cache_key] = puller_info
                break
        else:
            logger.debug("[add-watcher] no puller available, new one and add key:%s" % cache_key)
            key_list = self.process_mgr.list()
            key_list.append(cache_key)
            sys_os = platform.system()
            if sys_os == 'Windows':
                puller = Thread(target=self._do_pulling, args=(key_list, self.notify_queue))
                puller.setDaemon(True)
            else:
                puller = Process(target=self._do_pulling, args=(key_list, self.notify_queue))
                puller.daemon = True
            puller.start()
            self.puller_mapping[cache_key] = (puller, key_list)

    @synchronized_with_attr("pulling_lock")
    def remove_config_watcher(self, data_id, group, cb, remove_all=False):
        if not cb:
            raise NacosException("A callback function is needed.")
        data_id, group = process_common_config_params(data_id, group)
        if not self.puller_mapping:
            logger.warning("[remove-watcher] watcher is never started.")
            return
        cache_key = group_key(data_id, group, self.namespace)
        wl = self.watcher_mapping.get(cache_key)
        if not wl:
            logger.warning("[remove-watcher] there is no watcher on key:%s" % cache_key)
            return

        wrap_to_remove = list()
        for i in wl:
            if i.callback == cb:
                wrap_to_remove.append(i)
                if not remove_all:
                    break

        for i in wrap_to_remove:
            wl.remove(i)

        logger.info("[remove-watcher] %s is removed from %s, remove all:%s" % (cb.__name__, cache_key, remove_all))
        if not wl:
            logger.debug("[remove-watcher] there is no watcher for:%s, kick out from pulling" % cache_key)
            self.watcher_mapping.pop(cache_key)
            puller_info = self.puller_mapping[cache_key]
            puller_info[1].remove(cache_key)
            if not puller_info[1]:
                logger.debug("[remove-watcher] there is no pulling keys for puller:%s, stop it" % puller_info[0])
                self.puller_mapping.pop(cache_key)
                if isinstance(puller_info[0], Process):
                    puller_info[0].terminate()

    def _do_sync_req(self, url, headers=None, params=None, data=None, timeout=None, method="GET"):
        """执行请求"""
        if self.username and self.password:
            if not params:
                params = {}
            params.update({"username": self.username, "password": self.password})
        url = "?".join([url, urlencode(params)]) if params else url
        all_headers = self._get_common_headers(params, data)
        if headers:
            all_headers.update(headers)
        logger.debug(
            "[do-sync-req] url:%s, headers:%s, params:%s, data:%s, timeout:%s" % (
                url, all_headers, params, data, timeout))
        tries = 0
        while True:
            try:
                server_info = self.get_server()
                if not server_info:
                    logger.error("[do-sync-req] can not get one server.")
                    raise NacosRequestException("Server is not available.")
                address, port = server_info
                server = ":".join([address, str(port)])
                server_url = server
                if not server_url.startswith("http"):
                    server_url = "%s://%s" % ("http", server)
                if python_version_bellow("3"):
                    req = Request(url=server_url + url, data=urlencode(data).encode() if data else None,
                                  headers=all_headers)
                    req.get_method = lambda: method
                    ctx = ssl.create_default_context()
                    ctx.check_hostname = False
                    ctx.verify_mode = ssl.CERT_NONE
                else:
                    req = Request(url=server_url + url, data=urlencode(data).encode() if data else None,
                                  headers=all_headers, method=method)
                    ctx = ssl.SSLContext()
                # build a new opener that adds proxy setting so that http request go through the proxy
                if self.proxies:
                    proxy_support = ProxyHandler(self.proxies)
                    https_support = HTTPSHandler(context=ctx)
                    opener = build_opener(proxy_support, https_support)
                    resp = opener.open(req, timeout=timeout)
                else:
                    # for python version compatibility
                    if python_version_bellow("2.7.5"):
                        resp = urlopen(req, timeout=timeout)
                    else:
                        resp = urlopen(req, timeout=timeout, context=ctx)
                logger.debug("[do-sync-req] info from server:%s" % server)
                return resp
            except HTTPError as e:
                if e.code in [HTTPStatus.INTERNAL_SERVER_ERROR, HTTPStatus.BAD_GATEWAY,
                              HTTPStatus.SERVICE_UNAVAILABLE]:
                    logger.warning("[do-sync-req] server:%s is not available for reason:%s" % (server, e.msg))
                else:
                    raise
            except socket.timeout:
                logger.warning("[do-sync-req] %s request timeout" % server)
            except URLError as e:
                logger.warning("[do-sync-req] %s connection error:%s" % (server, e.reason))

            tries += 1
            if tries >= len(self.server_list):
                logger.error("[do-sync-req] %s maybe down, no server is currently available" % server)
                raise NacosRequestException("All server are not available")
            self.change_server()
            logger.warning("[do-sync-req] %s maybe down, skip to next" % server)

    def _do_pulling(self, cache_list, queue):
        cache_pool = dict()
        for cache_key in cache_list:
            cache_pool[cache_key] = CacheData(cache_key, self)

        while cache_list:
            unused_keys = set(cache_pool.keys())
            contains_init_key = False
            probe_update_string = ""
            for cache_key in cache_list:
                cache_data = cache_pool.get(cache_key)
                if not cache_data:
                    logger.debug("[do-pulling] new key added: %s" % cache_key)
                    cache_data = CacheData(cache_key, self)
                    cache_pool[cache_key] = cache_data
                else:
                    unused_keys.remove(cache_key)
                if cache_data.is_init:
                    contains_init_key = True
                data_id, group, namespace = parse_key(cache_key)
                probe_update_string += WORD_SEPARATOR.join(
                    [data_id, group, cache_data.md5 or "", self.namespace]) + LINE_SEPARATOR

            for k in unused_keys:
                logger.debug("[do-pulling] %s is no longer watched, remove from cache" % k)
                cache_pool.pop(k)

            logger.debug(
                "[do-pulling] try to detected change from server probe string is %s" % truncate(probe_update_string))
            headers = {"Long-Pulling-Timeout": int(self.pulling_timeout * 1000)}
            # if contains_init_key:
            #     headers["longPullingNoHangUp"] = "true"

            data = {"Listening-Configs": probe_update_string}

            changed_keys = list()
            try:
                resp = self._do_sync_req("/nacos/v1/cs/configs/listener", headers, None, data,
                                         self.pulling_timeout + 10, "POST")
                changed_keys = [group_key(*i) for i in parse_pulling_result(resp.read())]
                logger.debug("[do-pulling] following keys are changed from server %s" % truncate(str(changed_keys)))
            except NacosException as e:
                logger.error("[do-pulling] nacos exception: %s, waiting for recovery" % str(e))
                time.sleep(1)
            except Exception as e:
                logger.exception("[do-pulling] exception %s occur, return empty list, waiting for recovery" % str(e))
                time.sleep(1)

            for cache_key, cache_data in cache_pool.items():
                cache_data.is_init = False
                if cache_key in changed_keys:
                    data_id, group, namespace = parse_key(cache_key)
                    content = self.get_config(data_id, group)
                    cache_data.md5 = NacosClient.get_md5(content)
                    cache_data.content = content
                queue.put((cache_key, cache_data.content, cache_data.md5))

    @synchronized_with_attr("pulling_lock")
    def _init_pulling(self):
        if self.puller_mapping is not None:
            logger.info("[init-pulling] puller is already initialized")
            return
        self.puller_mapping = dict()
        self.notify_queue = Queue()
        self.callback_tread_pool = pool.ThreadPool(self.callback_thread_num)
        self.process_mgr = Manager()
        t = Thread(target=self._process_polling_result)
        t.setDaemon(True)
        t.start()
        logger.info("[init-pulling] init completed")

    def _process_polling_result(self):
        while True:
            cache_key, content, md5 = self.notify_queue.get()
            logger.debug("[process-polling-result] receive an event:%s" % cache_key)
            wl = self.watcher_mapping.get(cache_key)
            if not wl:
                logger.warning("[process-polling-result] no watcher on %s, ignored" % cache_key)
                continue

            data_id, group, namespace = parse_key(cache_key)
            plain_content = content

            params = {
                "data_id": data_id,
                "group": group,
                "namespace": namespace,
                "raw_content": content,
                "content": plain_content,
            }
            for watcher in wl:
                if not watcher.last_md5 == md5:
                    logger.debug(
                        "[process-polling-result] md5 changed since last call, calling %s with changed params: %s"
                        % (watcher.callback.__name__, params))
                    try:
                        self.callback_tread_pool.apply(watcher.callback, (params,))
                    except Exception as e:
                        logger.exception("[process-polling-result] exception %s occur while calling %s " % (
                            str(e), watcher.callback.__name__))
                    watcher.last_md5 = md5

    def _get_common_headers(self, params, data):
        headers = {}
        if self.auth_enabled:
            ts = str(int(time.time() * 1000))
            ak, sk = self.ak, self.sk

            headers.update({
                "Spas-AccessKey": ak,
                "timeStamp": ts,
            })
            sign_str = ""
            # in case tenant or group is null
            if not params and not data:
                return headers

            tenant = (params and params.get("tenant")) or (data and data.get("tenant"))
            group = (params and params.get("group")) or (data and data.get("group"))

            if tenant:
                sign_str = tenant + "+"
            if group:
                sign_str = sign_str + group + "+"

            if sign_str:
                sign_str += ts
                headers["Spas-Signature"] = base64.encodebytes(
                    hmac.new(sk.encode(), sign_str.encode(), digestmod=hashlib.sha1).digest()).decode().strip()
        return headers

    def _build_metadata(self, metadata, params):
        if metadata:
            if isinstance(metadata, dict):
                params["metadata"] = json.dumps(metadata)
            else:
                params["metadata"] = metadata


    def add_naming_instance(self, service_name, ip, port, cluster_name=None, weight=1.0, metadata=None,
                            enable=True, healthy=True, ephemeral=True,group_name=DEFAULT_GROUP_NAME):
        logger.info("[add-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s" % (
            ip, port, service_name, self.namespace))

        params = {
            "ip": ip,
            "port": port,
            "serviceName": service_name,
            "weight": weight,
            "enable": enable,
            "healthy": healthy,
            "clusterName": cluster_name,
            "ephemeral": ephemeral,
            "groupName": group_name
        }
        self._build_metadata(metadata, params)

        if self.namespace:
            params["namespaceId"] = self.namespace

        try:
            resp = self._do_sync_req("/nacos/v1/ns/instance", None, None, params, self.default_timeout, "POST")
            c = resp.read()
            logger.info("[add-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s, server response:%s" % (
                ip, port, service_name, self.namespace, c))
            return c == b"ok"
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                raise NacosException("Insufficient privilege.")
            else:
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[add-naming-instance] exception %s occur" % str(e))
            raise

    def remove_naming_instance(self, service_name, ip, port, cluster_name=None, ephemeral=True,group_name=DEFAULT_GROUP_NAME):
        logger.info("[remove-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s" % (
            ip, port, service_name, self.namespace))

        params = {
            "ip": ip,
            "port": port,
            "serviceName": service_name,
            "ephemeral": ephemeral,
            "groupName":group_name
        }

        if cluster_name is not None:
            params["clusterName"] = cluster_name

        if self.namespace:
            params["namespaceId"] = self.namespace

        try:
            resp = self._do_sync_req("/nacos/v1/ns/instance", None, None, params, self.default_timeout, "DELETE")
            c = resp.read()
            logger.info("[remove-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s, server response:%s" % (
                ip, port, service_name, self.namespace, c))
            return c == b"ok"
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                raise NacosException("Insufficient privilege.")
            else:
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[remove-naming-instance] exception %s occur" % str(e))
            raise

    def modify_naming_instance(self, service_name, ip, port, cluster_name=None, weight=None, metadata=None,
                               enable=None, ephemeral=True,group_name=DEFAULT_GROUP_NAME):
        logger.info("[modify-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s" % (
            ip, port, service_name, self.namespace))

        params = {
            "ip": ip,
            "port": port,
            "serviceName": service_name,
            "ephemeral": ephemeral,
            "groupName": group_name
        }

        if cluster_name is not None:
            params["clusterName"] = cluster_name

        if enable is not None:
            params["enable"] = enable

        if weight is not None:
            params["weight"] = weight

        self._build_metadata(metadata, params)

        if self.namespace:
            params["namespaceId"] = self.namespace

        try:
            resp = self._do_sync_req("/nacos/v1/ns/instance", None, None, params, self.default_timeout, "PUT")
            c = resp.read()
            logger.info("[modify-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s, server response:%s" % (
                ip, port, service_name, self.namespace, c))
            return c == b"ok"
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                raise NacosException("Insufficient privilege.")
            else:
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[modify-naming-instance] exception %s occur" % str(e))
            raise

    def list_naming_instance(self, service_name, clusters=None, namespace_id=None, group_name=None, healthy_only=False):
        """
        :param service_name:        服务名
        :param clusters:            集群名称            字符串，多个集群用逗号分隔
        :param namespace_id:        命名空间ID
        :param group_name:          分组名
        :param healthy_only:         是否只返回健康实例   否，默认为false
        """
        logger.info("[list-naming-instance] service_name:%s, namespace:%s" % (service_name, self.namespace))

        params = {
            "serviceName": service_name,
            "healthyOnly": healthy_only
        }

        if clusters is not None:
            params["clusters"] = clusters

        namespace_id = namespace_id or self.namespace
        if namespace_id:
            params["namespaceId"] = namespace_id

        group_name = group_name or 'DEFAULT_GROUP'
        if group_name:
            params['groupName'] = group_name

        try:
            resp = self._do_sync_req("/nacos/v1/ns/instance/list", None, params, None, self.default_timeout, "GET")
            c = resp.read()
            logger.info("[list-naming-instance] service_name:%s, namespace:%s, server response:%s" %
                        (service_name, self.namespace, c))
            return json.loads(c.decode("UTF-8"))
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                raise NacosException("Insufficient privilege.")
            else:
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[list-naming-instance] exception %s occur" % str(e))
            raise

    def get_naming_instance(self, service_name, ip, port, cluster_name=None):
        logger.info("[get-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s" % (ip, port, service_name,
                                                                                             self.namespace))

        params = {
            "serviceName": service_name,
            "ip": ip,
            "port": port,
        }

        if cluster_name is not None:
            params["cluster"] = cluster_name
            params["clusterName"] = cluster_name

        if self.namespace:
            params["namespaceId"] = self.namespace

        try:
            resp = self._do_sync_req("/nacos/v1/ns/instance", None, params, None, self.default_timeout, "GET")
            c = resp.read()
            logger.info("[get-naming-instance] ip:%s, port:%s, service_name:%s, namespace:%s, server response:%s" %
                        (ip, port, service_name, self.namespace, c))
            return json.loads(c.decode("UTF-8"))
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                raise NacosException("Insufficient privilege.")
            else:
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[get-naming-instance] exception %s occur" % str(e))
            raise

    def send_heartbeat(self, service_name, ip, port, cluster_name=None, weight=1.0, metadata=None, ephemeral=True,group_name=DEFAULT_GROUP_NAME):
        logger.info("[send-heartbeat] ip:%s, port:%s, service_name:%s, namespace:%s" % (ip, port, service_name,
                                                                                        self.namespace))
        beat_data = {
            "serviceName": service_name,
            "ip": ip,
            "port": port,
            "weight": weight,
            "ephemeral": ephemeral

        }

        if cluster_name is not None:
            beat_data["cluster"] = cluster_name

        if metadata is not None:
            if isinstance(metadata, str):
                beat_data["metadata"] = json.loads(metadata)
            else:
                beat_data["metadata"] = metadata

        params = {
            "serviceName": service_name,
            "beat": json.dumps(beat_data),
            "groupName": group_name
        }

        if self.namespace:
            params["namespaceId"] = self.namespace

        try:
            resp = self._do_sync_req("/nacos/v1/ns/instance/beat", None, params, None, self.default_timeout, "PUT")
            c = resp.read()
            logger.info("[send-heartbeat] ip:%s, port:%s, service_name:%s, namespace:%s, server response:%s" %
                        (ip, port, service_name, self.namespace, c))
            return json.loads(c.decode("UTF-8"))
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                raise NacosException("Insufficient privilege.")
            else:
                raise NacosException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[send-heartbeat] exception %s occur" % str(e))
            raise

    def subscribe(self,
                  listener_fn, listener_interval=7, *args, **kwargs):
        """
        reference at `/nacos/v1/ns/instance/list` in https://nacos.io/zh-cn/docs/open-api.html
        :param listener_fn           监听方法，可以是元组，列表，单个监听方法
        :param listener_interval     监听间隔，在 HTTP 请求 OpenAPI 时间间隔
        :return:
        """
        service_name = kwargs.get("service_name")
        if not service_name:
            if len(args) > 0:
                service_name = args[0]
            else:
                raise NacosException("`service_name` is required in subscribe")
        self.subscribed_local_manager.add_local_listener(key=service_name, listener_fn=listener_fn)

        #  判断是否是第一次订阅调用
        class _InnerSubContext(object):
            first_sub = True

        def _compare_and_trigger_listener():
            #  invoke `list_naming_instance`
            latest_res = self.list_naming_instance(*args, **kwargs)
            latest_instances = latest_res['hosts']
            #  获取本地缓存实例
            local_service_instances_dict = self.subscribed_local_manager.get_local_instances(service_name)
            #  当前本地没有缓存，所有都是新的实例
            if not local_service_instances_dict:
                if not latest_instances or len(latest_instances) < 1:
                    #  第一次订阅调用不通知
                    if _InnerSubContext.first_sub:
                        _InnerSubContext.first_sub = False
                        return
                for instance in latest_instances:
                    slc = SubscribedLocalInstance(key=service_name, instance=instance)
                    self.subscribed_local_manager.add_local_instance(slc)
                    #  第一次订阅调用不通知
                    if _InnerSubContext.first_sub:
                        _InnerSubContext.first_sub = False
                        return
                    self.subscribed_local_manager.do_listener_launch(service_name, Event.ADDED, slc)
            else:
                local_service_instances_dict_copy = local_service_instances_dict.copy()
                for instance in latest_instances:
                    slc = SubscribedLocalInstance(key=service_name, instance=instance)
                    local_slc = local_service_instances_dict.get(slc.instance_id)
                    # 本地不存在实例缓存
                    if local_slc is None:
                        self.subscribed_local_manager.add_local_instance(slc)
                        self.subscribed_local_manager.do_listener_launch(service_name, Event.ADDED, slc)
                    # 本地存在实例缓存
                    else:
                        local_slc_md5 = local_slc.md5
                        local_slc_id = local_slc.instance_id
                        local_service_instances_dict_copy.pop(local_slc_id)
                        # 比较md5,存在实例变更
                        if local_slc_md5 != slc.md5:
                            self.subscribed_local_manager.remove_local_instance(local_slc).add_local_instance(slc)
                            self.subscribed_local_manager.do_listener_launch(service_name, Event.MODIFIED, slc)
                #  still have instances in local marked deleted
                if len(local_service_instances_dict_copy) > 0:
                    for local_slc_id, slc in local_service_instances_dict_copy.items():
                        self.subscribed_local_manager.remove_local_instance(slc)
                        self.subscribed_local_manager.do_listener_launch(service_name, Event.DELETED, slc)

        timer_name = 'service-subscribe-timer-{key}'.format(key=service_name)
        subscribe_timer = NacosTimer(name=timer_name,
                                     interval=listener_interval,
                                     fn=_compare_and_trigger_listener)
        subscribe_timer.scheduler()
        self.subscribe_timer_manager.add_timer(subscribe_timer)

    def unsubscribe(self, service_name, listener_name=None):
        """
        remove listener from subscribed  listener manager
        :param service_name:    service_name
        :param listener_name:   listener name
        :return: 
        """
        listener_manager = self.subscribed_local_manager.get_local_listener_manager(key=service_name)
        if not listener_manager:
            return
        if listener_name:
            listener_manager.remove_listener(listener_name)
            return
        listener_manager.empty_listeners()

    def stop_subscribe(self):
        """
        stop subscribe timer scheduler
        :return: 
        """
        self.subscribe_timer_manager.stop()


if DEBUG:
    NacosClient.set_debugging()





import os
import requests



class HaiModel(object):

    @staticmethod
    def ensure_api_key(api_key):
        if api_key is None:
            api_key = os.environ.get("HEPAI_API_KEY", None)
        if api_key is None:
            raise ValueError(f"""            
The HepAI API-KEY is required. Please set the environment variable `HEPAI_API_KEY` via `export HEPAI_API_KEY=xxx`.
Alternatively, it can be provided by passing in the `api_key` parameter when calling the method.
""")
        return api_key

    @staticmethod
    def list(**kwargs):
        """List all models.

        :return: The list of models.
        """  
        host = kwargs.get("host", "chat.ihep.ac.cn")
        port = kwargs.get("port", None)
        if port:
            url = f"http://{host}:{port}"
        else:
            url = f'https://{host}'

        ret = requests.post(
            f"{url}/list_models",
            )
        return ret.json()
    
    @staticmethod
    def all_workers_info(**kwargs):
        models = HaiModel.list()
        api_key = kwargs.pop("api_key", None)
        api_key = HaiModel.ensure_api_key(api_key)
        host = kwargs.get("host", "chat.ihep.ac.cn")
        port = kwargs.get("port", None)
        assert api_key, f'api_key is required.'
        if port:
            url = f"http://{host}:{port}"
        else:
            url = f'https://{host}'
        headers={"Authorization": f"Bearer {api_key}"}
        all_info = dict()
        for model_name in models['models']:
            response = requests.post(
                f"{url}/get_worker_info",
                headers=headers,
                json={
                    "model": model_name,
                },
            )
            worker_info = response.json()['info']
            workers = worker_info[model_name]  # 一个模型可能有多个worker
            workers_info = dict()
            for one_worker in workers:
                address = one_worker.pop('address')
                
                workers_info[address] = one_worker
            # all_info.update(worker_info)
            all_info[model_name] = workers_info
            # print(res.json())
        return all_info

    
if __name__ == '__main__':
    ret = HaiModel.list()
    print(ret)



import logging
from os import listdir
from os.path import splitext
from pathlib import Path

import numpy as np
import torch
from PIL import Image
from torch.utils.data import Dataset


class BasicDataset(Dataset):
    def __init__(self, images_dir: str, masks_dir: str, scale: float = 1.0, mask_suffix: str = ''):
        self.images_dir = Path(images_dir)
        self.masks_dir = Path(masks_dir)
        assert 0 < scale <= 1, 'Scale must be between 0 and 1'
        self.scale = scale
        self.mask_suffix = mask_suffix

        self.ids = [splitext(file)[0] for file in listdir(images_dir) if not file.startswith('.')]
        if not self.ids:
            raise RuntimeError(f'No input file found in {images_dir}, make sure you put your images there')
        logging.info(f'Creating dataset with {len(self.ids)} examples')

    def __len__(self):
        return len(self.ids)

    @staticmethod
    def preprocess(pil_img, scale, is_mask):
        w, h = pil_img.size
        newW, newH = int(scale * w), int(scale * h)
        assert newW > 0 and newH > 0, 'Scale is too small, resized images would have no pixel'
        pil_img = pil_img.resize((newW, newH), resample=Image.NEAREST if is_mask else Image.BICUBIC)
        img_ndarray = np.asarray(pil_img)

        if not is_mask:
            if img_ndarray.ndim == 2:
                img_ndarray = img_ndarray[np.newaxis, ...]
            else:
                img_ndarray = img_ndarray.transpose((2, 0, 1))

            img_ndarray = img_ndarray / 255

        return img_ndarray

    @staticmethod
    def load(filename):
        ext = splitext(filename)[1]
        if ext == '.npy':
            return Image.fromarray(np.load(filename))
        elif ext in ['.pt', '.pth']:
            return Image.fromarray(torch.load(filename).numpy())
        else:
            return Image.open(filename)

    def __getitem__(self, idx):
        name = self.ids[idx]
        mask_file = list(self.masks_dir.glob(name + self.mask_suffix + '.*'))
        img_file = list(self.images_dir.glob(name + '.*'))

        assert len(img_file) == 1, f'Either no image or multiple images found for the ID {name}: {img_file} in {self.images_dir}'
        assert len(mask_file) == 1, f'Either no mask or multiple masks found for the ID {name}: {mask_file} in {self.masks_dir}'
        mask = self.load(mask_file[0])
        img = self.load(img_file[0])

        assert img.size == mask.size, \
            f'Image and mask {name} should be the same size, but are {img.size} and {mask.size}'

        img = self.preprocess(img, self.scale, is_mask=False)
        mask = self.preprocess(mask, self.scale, is_mask=True)

        return {
            'image': torch.as_tensor(img.copy()).float().contiguous(),
            'mask': torch.as_tensor(mask.copy()).long().contiguous()
        }


class CarvanaDataset(BasicDataset):
    def __init__(self, images_dir, masks_dir, scale=1):
        super().__init__(images_dir, masks_dir, scale, mask_suffix='')




import hai
from hai import AbstractModule, AbstractInput, AbstractOutput, AbstractQue
from hai import MODULES, SCRIPTS, IOS, Config



@hai.MODULES.register_module(name='YOLOv5')
class YOLOv5v6(AbstractModule):
    name = 'YOLOv5'
    description = '2020 SOTA 目标检测算法YOLOv5-v6'

    def __init__(self, ):
        super().__init__()
        # self._config = _config

    # def get_cfg(self):
        # return self.cfg

    


model = dict(
    type='deepsort',
    weights="~/weights/xsensing/deepsort_ckpt.t7",
    use_cuda=False,
    multi_tracker=1,  # 跟踪器数目
)

output_stream = dict(
    type='uaii_deepsort_output',
    que=dict(
        enabled=True,
        maxlen=5,
        wait=True
    ),
)



import hai


@hai.MODULES.register_module(name='ResNet')
class ResNet(hai.AbstractModule):
    name = 'ResNet'
    description = 'ResNet网络'

    def __init__(self, *args, **kwargs):
        super(ResNet, self).__init__(*args, **kwargs)
        self._init()

    def _init(self):
        pass

    def __call__(self, *args, **kwargs):
        pass

    def single_plot(self, img, preds, names):
        """
        单张图片绘制
        :param img:
        :param bboxes:
        :param scores:
        :param classes:
        :return:
        """
        # print('bb', bboxes)
        # print(f'scores: {scores}')
        # print(f'classes: {classes}')
        bboxes = preds[:, :4]
        scores = preds[:, 4]
        classes_idx = [int(x) for x in preds[:, 5]]
        classes = [names[int(x)] for x in classes_idx]
        colors = [(204, 0, 0), (0, 130, 255), (0, 204, 0), (0, 0, 204), (255, 130, 0)]

        for i, bbox in enumerate(bboxes):
            score = scores[i]
            cls_idx = classes_idx[i]
            cls = names[cls_idx]


"""
uaii general
"""
import damei as dm
import numpy as np

from ..datasets.dataset_utils import get_file, extract_archive


def single_plot(result, orig_img, target_names):
    """
    把检测结果绘制到原始图像上。
    :param result: ndarray [n, 6] 或者 [n, 9],
    :param orig_img:
    :param target_names: 目标类别list
    :return:
    """
    if result is None or len(result.shape) == 1:
        return orig_img
    else:
        out_img = np.copy(orig_img)
        # result  [num_obj, 10]，10: xyxy cls tid trace keypoints kp_score proposal_score
        # print('')
        # print(result.shape)
        shape = result.shape
        # print(shape)
        if shape[1] == 6:
            format = 'seyolov5'
        else:
            raise NotImplementedError(f'仅支持seyolov5的输出格式，{shape}')

        for i, target in enumerate(result):
            if format == 'seyolov5':
                bbox_xyxy = target[:4]

                target_cls = target[5]
                confidence = target[4]
                target_name = target_names[int(target_cls)]
                label = f"{target_name} {confidence:.2f}"
                trace = None
                status = None
                keypoints = None
                kp_score = None
            else:
                bbox_xyxy = target[:4]
                target_cls = target[4]
                tid = target[5]
                trace = target[6]
                keypoints = target[7]
                kp_score = target[8]
                pro_socre = target[9]

                status_idx = target[10]
                status_name = target[11]
                status_score = target[12]

                status_topx = 2
                if status_idx is None:
                    status = None
                else:
                    status = [f'{status_score[k]:.2f} {x}' for k, x in enumerate(status_name[:status_topx:])]

                target_name = target_names[target_cls]
                status = status if target_name == 'person' else None

                # print(
                # 	f'\nobj_idx: {i} target: {target_name} tracking_id: {tid}\nidx  : {status_idx} '
                # 	f'\nname : {status_name} \nscore: {status_score}')

                first_status = status_name[0] if status_name is not None else ''
                label = f"{tid:<2} {target_name} {first_status}"
            # exit()
            # status = person[11] if person[11] is not None else 'unk'
            out_img = dm.general.plot_one_box_trace_pose_status(
                bbox_xyxy, out_img,
                label=label, color=None, trace=trace, status=status,
                line_thickness=4, keypoints=keypoints, kp_score=kp_score)
        return out_img

    
def dict_match_item(all_dict, name, version):
    """
    从字典中匹配key和value
    :param dict:
    :param key:
    :param value:
    :return:
    """
    keys = list(all_dict.keys())
    name = name.lower()
    macthed = [x for x in keys if name in x.lower()]
    if len(macthed) == 0:
        return None
    else:
        if version is None:
            version = latest2determined(all_dict, name)
        for x in macthed:
            if all_dict[x]['version'] == version:
                return all_dict[x]
        return None
    

def latest2determined(all_dict, name):
    """
    从字典中找到最新的指定版本
    return: the latest version, i.e. 1.1 or 1.2
    """
    keys = list(all_dict.keys())
    name = name.lower()
    macthed = [x for x in keys if name in x.lower()]
    if len(macthed) == 0:
        return None
    else:
        versions = [all_dict[x]['version'] for x in macthed]
        version = sorted(versions)[-1]
        return version






"""
Worker类别，具备基本和示例功能
"""

import time
import json
import time
from typing import List, Union
import threading
import uuid
import traceback
import argparse
# import dataclasses
from dataclasses import dataclass, field
import os, sys
from pathlib import Path

from fastapi import FastAPI, Request, BackgroundTasks
from fastapi.responses import StreamingResponse
import requests
import uvicorn

import damei as dm 
here = Path(__file__).parent

try:
    from hai.uaii.worker.utils import auto_port, auto_worker_address, pretty_print_semaphore
    from hai.uaii.worker.base_worker_model import BaseWorkerModel
except:
    sys.path.append(str(here.parent.parent.parent))
    from hai.uaii.worker.utils import auto_port, auto_worker_address, pretty_print_semaphore
    from hai.uaii.worker.base_worker_model import BaseWorkerModel


GB = 1 << 30

logger = dm.get_logger('base_worker.py')
global_counter = 0
model_semaphore = None
WORKER_HEART_BEAT_INTERVAL = 30


def heart_beat_worker(controller):
    while True:
        time.sleep(WORKER_HEART_BEAT_INTERVAL)
        controller.send_heart_beat()



@dataclass
class WorkerArgs:
    host: str = "0.0.0.0"  # worker的地址，0.0.0.0表示外部可访问，127.0.0.1表示只有本机可访问
    port: str = "auto"  # 默认从42902开始
    controller_address: str = "http://chat.ihep.ac.cn:42901"  # 控制器的地址
    worker_address: str = "auto"  # 默认是http://<ip>:<port>
    limit_model_concurrency: int = 5  # 限制模型的并发请求
    stream_interval: float = 0.  # 额外的流式响应间隔
    no_register: bool = False  # 不注册到控制器
    premissions: str = 'group: all'  # 模型的权限授予，分为用户和组，用;分隔


class Worker:
    """
    Worker class, with basic and example functions
    """
    def __init__(self, 
                 controller_addr: str,  # controller的地址
                 worker_addr: str,  # 本worker的地址
                 worker_id: str = None,  # worker的uuid
                 model=None,  # 模型
                 limit_model_concurrency: int = 5,  # 模型并发数
                 stream_interval: int = 2,  # stream的间隔
                 no_register: bool = False,  # 是否注册到controller
                 premissions: any = None,  # 权限
                 **kwargs
                 ):
        
        self.controller_addr = controller_addr
        self.worker_addr = worker_addr
        self.worker_id = worker_id if worker_id else str(uuid.uuid4())[:6]
        self.limit_model_concurrency = limit_model_concurrency
        self.stream_interval = stream_interval
        self.no_register = no_register
        self._premissions = self._init_premissions(premissions)

        self._model = model or BaseWorkerModel()

        if not no_register:
            self.register_to_controller()
            self.heart_beat_thread = threading.Thread(
                target=heart_beat_worker, args=(self,))
            self.heart_beat_thread.start()

    def _init_premissions(self, premissions):
        """worker授予用户或者组的权限"""
        prems = dict()
        if premissions is None:
            pass
        elif isinstance(premissions, str):
            """user: <user1>; user: <user2>; group: <group1>, ..."""
            for a in premissions.split(';'):
                user_or_group, name = a.split(':')
                user_or_group = user_or_group.strip()
                name = name.strip()
                assert user_or_group in ['user', 'group']
                prems[user_or_group] = name
        elif isinstance(premissions, dict):
            prems = premissions
        else:
            raise ValueError(f"premissions should be str or dict, but got {type(premissions)}")
        return prems

    def check_model(self):
        # 测试是否有inference函数
        assert hasattr(self.model, 'inference'), f"Model {self.model_name} has no inference function"

    @property
    def premissions(self):
        return self._premissions
    
    @property
    def model_name(self):
        if self.model is None:
            return None
        return self.model.name

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._model = model    
    
    def register_to_controller(self):
        logger.info(f'Register model "{self.model_name}" to controller.')
        url = self.controller_addr + "/register_worker"
        data = {
            "worker_name": self.worker_addr,
            "check_heart_beat": True,
            "worker_status": self.get_status(),
            }
        r = requests.post(url, json=data)
        assert r.status_code == 200

    def send_heart_beat(self):
        logger.info(f"Send heart beat. Models: {[self.model_name]}. "
                    f"Semaphore: {pretty_print_semaphore(model_semaphore)}. "
                    f"global_counter: {global_counter}")

        url = self.controller_addr + "/receive_heart_beat"

        while True:
            try:
                ret = requests.post(url, json={
                    "worker_name": self.worker_addr,
                    "queue_length": self.get_queue_length()}, timeout=5)
                exist = ret.json()["exist"]
                break
            except requests.exceptions.RequestException as e:
                logger.error(f"heart beat error: {e}")
            time.sleep(5)

        if not exist:
            self.register_to_controller()

    def get_queue_length(self):
        if model_semaphore is None:
            return 0
        else:
            _value = model_semaphore._value
            _num_waiters = len(model_semaphore._waiters) if model_semaphore._waiters is not None else 0
            return self.limit_model_concurrency - _value + _num_waiters

    def get_status(self):
        try:
            misc = self.model.get_misc()
        except:
            misc = None
        return {
            "model_names": [self.model_name],
            "speed": 1,
            "queue_length": self.get_queue_length(),
            "premissions": self.premissions,
            "misc": misc,
        }

    def _sleep(self, seconds):
        if float(seconds) == 0. or seconds is None:
            return
        if seconds < 0:
            raise ValueError(f"stream_interval should be positive, but got {seconds}")
        if seconds > 3:
            raise ValueError(f"stream_interval should be less than 3, but got {seconds}")
        time.sleep(seconds)

    
    def get_generator_content_and_verify_one(self, generator):
        """
        从生成器中获取内容，并验证是否只有一个元素
        """
        try:
            content = next(generator)
            print('content: ', content)
        except StopIteration as e:
            print(f'Generator is empty: {e} {type(e)}')
            return e
            # raise ValueError(f"Generator is empty: {e} {type(e)}")
        try:
            next(generator)
            raise ValueError("Generator should only have one element")
        except StopIteration:
            pass
        return content

    
    def generate_stream_gate(self, **params):
        """
        生成流式响应的门
        :param params:
                model: 模型名
                其他参数
        :return:
        """
        
        stream = params.get("stream", False)
        stream_interval = params.get("stream_interval", self.stream_interval)

        def yield_data(output):
            if isinstance(output, int):
                output = str(output)
                yield output.encode() + b"\0"
            elif isinstance(output, str):
                for x in output:
                    yield x.encode() + b"\0"
            elif isinstance(output, bytes):
                yield output + b"\0"
            elif isinstance(output, list):
                output = str(output)
                yield output.encode() + b"\0"
            elif isinstance(output, dict):
                yield json.dumps(output).encode() + b"\0"
            else:
                raise ValueError(f'output type {type(output)} not supported')

        try:
            output = self.model.inference(**params)
            if not stream:
                # print(f'output: {output}')
                new_output = dict()
                new_output["message"] = output
                new_output["status_code"] = 42901
                yield json.dumps(new_output).encode()
                return
            else:
                if isinstance(output, (int, str, bytes, list, dict)):
                    yield from yield_data(output)
                    self._sleep(stream_interval)
                # 如果是python的generator
                elif hasattr(output, '__next__'):
                    for x in output:
                        yield from yield_data(x)
                        self._sleep(stream_interval)
                else:
                    raise ValueError(f'output type {type(output)} not supported')
                yield "[DONE]".encode() + b"\0"

        except Exception as e:
            error_info = f'{type(e)} {e}'
            logger.error(f'error_info: {error_info}\n {traceback.format_exc()}')
            ret = {
                "message": error_info,
                "status_code": 42904, 
            }
            yield json.dumps(ret).encode() + b"\0"

    def train(self, **kwargs):
        ret = self.model.train(**kwargs)
        return ret


from fastapi import FastAPI, Request, BackgroundTasks
from fastapi.responses import StreamingResponse
import asyncio


class WorkerAPP(FastAPI):

    def __init__(self):
        super().__init__()
        self._worker = None

    @property
    def worker(self):
        if self._worker is None:
            raise ValueError(
                f"worker is not initialized, \
                please call `WorkerAPP.worker = BaseWorker(**kwargs)` first")
        return self._worker
    
    @worker.setter
    def worker(self, worker):
        self._worker = worker
        
app = WorkerAPP()  # It's a FastAPI instance


def release_model_semaphore():
    model_semaphore.release()

@app.post("/worker_generate_stream")
async def generate_stream(request: Request):
    global model_semaphore, global_counter
    global_counter += 1
    params = await request.json()

    if model_semaphore is None:
        model_semaphore = asyncio.Semaphore(
            app.worker.limit_model_concurrency)
        # 用于设置模型的并发请求，并发达到该值时，后续请求会被阻塞，直到之前的请求完成释放了锁
    await model_semaphore.acquire()
    generator = app.worker.generate_stream_gate(**params)
    background_tasks = BackgroundTasks()  # 背景任务
    background_tasks.add_task(release_model_semaphore)  # 释放锁
    return StreamingResponse(generator, background=background_tasks)

@app.post("/train")
async def train(request: Request):
    global model_semaphore, global_counter
    global_counter += 1
    params = await request.json()
    if model_semaphore is None:
        model_semaphore = asyncio.Semaphore(
            app.worker.limit_model_concurrency
        )
    await model_semaphore.acquire()
    result = app.worker.train(**params)
    background_tasks = BackgroundTasks()  # 背景任务
    background_tasks.add_task(release_model_semaphore)  # 释放锁
    return result


@app.post("/worker_get_status")
async def get_status(request: Request):
    return app.worker.get_status()


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--host", type=str, default="127.0.0.1")
    parser.add_argument("--port", type=str, default='auto')
    parser.add_argument("--controller-address", type=str,
        default="http://127.0.0.1:42901")
    parser.add_argument("--worker-address", type=str,
        default="auto")
    parser.add_argument("--limit-model-concurrency", type=int, default=5, help="限制模型的并发请求")
    parser.add_argument("--stream-interval", type=int, default=0., help="额外的流式响应间隔")
    parser.add_argument("--no-register", action="store_true", help="不注册到控制器")
    parser.add_argument("--premissions", type=str, default='group: all', 
        help="模型权限授予谁,写法：'user: <user1>; user: <user2>; group: <group1>, ...'")
    args = parser.parse_args()
    logger.info(f"Args: {args}")
    return args

    
def run_worker(model=None, worker_args=None, daemon=False, test=False, **kwargs):
    # args = get_args() if args is None else args
    # import transformers
    # args = transformers.HfArgumentParser((WorkerArgs, )).parse_args_into_dataclasses()
    args = worker_args or WorkerArgs()
    # print(f'worker args: {args}')
    args.port = auto_port(args.port, start=42902)
    args.worker_address = auto_worker_address(args.worker_address, args.host, args.port)
    if test:
        args.controller_address = "http://chat.ihep.ac.cn:4444"

    # 更新
    for k, v in kwargs.items():
        if hasattr(args, k):
            setattr(args, k, v)

    # logger.info(f'{args}')

    worker  = Worker(
        args.controller_address,
        args.worker_address,
        model=model,
        limit_model_concurrency=args.limit_model_concurrency,
        stream_interval=args.stream_interval,
        no_register=args.no_register,
        premissions=args.premissions
        )
    
    app.worker = worker

    logger.info(f"Controller address: {args.controller_address}")

    if daemon:
        logger.info(f"Daemon Worker address: {args.worker_address}")
        import threading
        t = threading.Thread(target=uvicorn.run, args=(app,), kwargs={
            "host": args.host, "port": args.port, "log_level": "info"
            })
        t.setDaemon(True)
        t.start()
    else:
        uvicorn.run(app, host=args.host, port=args.port, log_level="info")
    return args.worker_address


class WorkerWarper:

    @staticmethod
    def start(daemon=False, **kwargs):
        """
        运行一个hepai的worker
        :param model: BaseWorkerModel = None, 模型
        :param worker_args: WorkerArgs = None, worker的参数，可以由WorkerArgs类生成，会被下面的参数覆盖
        :praam daemon: bool = False 是否以守护进程的方式运行
        :param host: str = "0.0.0.0"  # worker的地址，0.0.0.0表示外部可访问，127.0.0.1表示只有本机可访问
        :param port: str = "auto"  # 默认从42902开始
        :param controller_address: str = "http://chat.ihep.ac.cn:42901"  # 控制器的地址
        :param worker_address: str = "auto"  # 默认是http://<ip>:<port>
        :param limit_model_concurrency: int = 5  # 限制模型的并发请求
        :param stream_interval: float = 0.  # 额外的流式响应间隔
        :param no_register: bool = False  # 不注册到控制器
        :param premissions: str = 'group: all'  # 模型的权限授予，分为用户和组，用;分隔
        """
        return run_worker(daemon=daemon, **kwargs)


if __name__ == "__main__":
    run_worker(test=False)
    
    











import os, sys
import shutil

import argparse

import damei as dm
import hai
from ..utils.config_loader import PyConfigLoader as Config
from ...testor import Testor
from .cli_functions import CLIFunctions

logger = dm.getLogger('hai_cli')


def run():
    args = argparse.ArgumentParser()
    args.add_argument('mode', type=str, nargs='?', default=None, help='Operation, such as: list, download, train, eval, deploy, etc.')
    args.add_argument('sub_mode', type=str, nargs='?', default=None, help='Target, such as: <model_name>, <dataset_name>, remote, etc.')
    args.add_argument('sub_sub_mode', type=str, nargs='?', default=None, help='Additional parameter.')
    args.add_argument('-V', '--version', action='store_true', help='show version')
    args.add_argument('-f', '--force', action='store_true', help='force to run (if the api exists, clear it and init again)')
    opt = args.parse_args()

    cli = CommandLineInterface()
    cli(opt) 


class CommandLineInterface(CLIFunctions):
    def __init__(self, uaii=None, config=None):
        self.uaii = uaii if uaii is not None else hai.UAII()
        self.config = config if config else hai.config # this is the hai config
        self.default_model = None  # if run hai command in a folder containing a model, then set the model as default model
        self.opt = None
        self.testor = Testor()
    
    def _init_opt(self, opt):
        # is cwd not in sys.path, then add it
        cwd = os.getcwd()
        if cwd not in sys.path:
            sys.path.append(cwd)

        is_in_one_module = hai.config.API_FOLD_NAME in [x for x in os.listdir('.') if os.path.isdir(x)]
        if is_in_one_module:
            self.default_model = os.path.basename(os.getcwd())
        self.opt = opt
        return opt

    def __call__(self, opt, **kwargs):
        # print('opt: ', opt)
        opt = self._init_opt(opt)
        mode = opt.mode
        if mode == 'init':
            self._init_a_module(opt)  
        elif mode == 'list':
            self._deal_with_list(opt, **kwargs)
        elif mode == 'download':
            self._deal_with_download(opt, **kwargs)
        elif mode == 'model':
            self._deal_model_mode(opt)
        elif mode == 'models':
            self._list_all_modules(module=True)
        elif mode == 'weights':
            self._deal_weights_mode(opt)
        elif mode == 'datasets':
            self._deal_datasets_mode(opt)
        elif mode == 'test':
            self._deal_test_mode(opt)
        elif mode == 'train':
            self._deal_train_mode(opt)
        elif mode == 'version' or opt.version:
            self._show_version()
        elif mode == 'start':
            self._deal_with_start(**kwargs)
        elif mode is None:
            self._show_version()
            print(f'Please use "{hai.__appname__} -h" to see help.')
        else:
            exists_model_names = self.uaii.module_names
            if mode in exists_model_names:
                self._deal_exists_model(opt)
            else:
                raise NotImplementedError(f'Not implemented mode: {mode}')

    def _deal_exists_model(self, opt):
        model_name = opt.mode
        sub_mode = opt.sub_mode
        if sub_mode == 'configs':
            if not self.uaii.model:
                self.uaii.load_model(model_name)
            model = self.uaii.model
            assert self.uaii.model_name == model_name, f'Error: model_name: {model_name} != self.uaii.model_name: {self.uaii.model_name}'
            print(f'{model_name}: {model.config}')
            # self._list_all_modules(module=True)
        else:
            raise NotImplementedError(f'Not implemented sub_mode: {sub_mode}')

    def _deal_train_mode(self, opt):
        model_name = opt.sub_mode if opt.sub_mode else self.default_model
        # print(os.getcwd())
        model = self.uaii.load_model(model_name)
        model.train()

    
    def _deal_test_mode(self, opt):
        model_name = opt.sub_mode if opt.sub_mode else self.default_model
        ret = self.testor.test_module(model_name)
        # assert ret, f'Error: current test failed for module: {model_name}'

    def _deal_model_mode(self, opt):
        sub_mode = opt.sub_mode
        if sub_mode == 'list':
            self._list_all_modules(module=True)
        else:  # hai model 
            # current_model_name = self.uaii._model_name
            # print(f'Current_model_name: {current_model_name}')
            raise NotImplementedError(f'Not implemented sub_mode: {sub_mode}')
        
    def _deal_weights_mode(self, opt):
        opt = Config.from_argparse(opt)
        opt = opt.to_dict()
        _ = opt.pop('mode')
        model_name = opt.pop('sub_mode', None)
        weights = self.uaii.list_weights(model_name=model_name, **opt)
        print(f'Total {len(weights)} weights: {weights}')

    def _deal_datasets_mode(self, opt):
        sub_mode = opt.sub_mode
        ss_mode = opt.sub_sub_mode
        if sub_mode is None:  # list dataset in hai.DATASETS_ROOT
            info = self.uaii.list_datasets(sub_mode=sub_mode)
        # hai datasets list
        elif sub_mode == 'list':
            info = self.uaii.list_datasets(sub_mode=sub_mode, ret_fmt='dict')
            info = f'Total {len(info.keys())} datasets: {list(info.keys())}'
        elif sub_mode == 'info':
            info = self.uaii.list_datasets(sub_mode=sub_mode, ret_fmt='dict')
            info = dm.misc.dict2info(info) + \
            f'Total {len(info.keys())} datasets: {list(info.keys())}'
        elif sub_mode == 'download':  # download dataset
            info = self.uaii.download_dataset(name=ss_mode)
        else:
            if ss_mode == 'info':  # like: hai datasets SFID info
                dset_name = sub_mode.lower()
                ret = self.uaii.list_datasets(sub_mode='list', ret_fmt='dict')
                avail_datasets = list(ret.keys())
                if 'latest' in dset_name:  # like: hai datasets SFID:latest info
                    dset_name = dset_name.split(':')[0]
                    match = [x for x in avail_datasets if dset_name in x.lower()]
                    match = list(reversed(match))
                    if len(match) > 0:
                        versions = [ret[x]['version'] for x in match]
                        # print(f'Available versions: {versions}')
                        latest_version = sorted(versions)[-1]
                        dset_name = f'{dset_name}:{latest_version}'
                match = [x for x in avail_datasets if dset_name in x.lower()]
                # print(match)
                if len(match) > 0: 
                    info = {}
                    for x in match:
                        info[x] = ret[x]
                    info = f'{sub_mode} dataset info:\n'+dm.misc.dict2info(info)
                else:
                    raise NotImplementedError(f'Not implemented dataset: {sub_mode}, opt: {opt}')
            else:
                raise ValueError(f'sub_mode: {sub_mode} not supported. opt: {opt}')
        print(info)
        
        
    def _init_a_module(self, opt):
        cfg = self.config
        api_fold_name = cfg.API_FOLD_NAME  # i.e. 'hai_api
        root_path = cfg.ROOT_PATH  # i.e. '/home/xxx/VSProjects/hai'

        if os.path.exists(api_fold_name):
            if not opt.force:
                logger.warning(f'API folder "{api_fold_name}" exists, please use "-f" to force to init again')
                return
            else:  # force
                pass
        else:                                                                                                                                                                                                                                                                                                                                                                                           
            os.mkdir(api_fold_name)

        cwd = os.getcwd()
        project_name = os.path.basename(cwd).lower()

        # Copy template files to api_fold
        shutil.copy(f'{root_path}/hai/apis/templates/README_template.md', 
                    f'{api_fold_name}/README.md')
        shutil.copy(f'{root_path}/hai/apis/templates/register_module_template.py', 
                    f'{api_fold_name}/{project_name}_api.py')
        shutil.copy(f'{root_path}/hai/apis/templates/init_template.py', 
                    f'{api_fold_name}/__init__.py')
        os.system(f'echo "\nfrom .{project_name}_api import *" >> {api_fold_name}/__init__.py')
        logger.info(f'Create API folder and files successfully: "{os.getcwd()}/{api_fold_name}"')


    def _show_version(self):
        data = dict()
        ver = hai.__version__ + '-' + hai.__version_suffix__
        data[f'{hai.__appname__.upper()} Version'] = ver
        # data['Author'] = hai.__author__
        data['URL'] = hai.__url__
        data['Contact'] = f'For any suggestions or demands, please email: {hai.__email__}'
        # print(f'HAI Version: {hai.__version__}, Author: {hai.__author__} @ {hai.__affiliation__}, Email: {hai.__email__}')
        info = dm.misc.dict2info(data)
        info = info[:-1] if info.endswith('\n') else info  # remove last \n
        print(info)

    def run(self):
        pass


